#include "/prelude/core.glsl"

/* SMAA Neighborhood Blending & RCA Sharpening */

layout(local_size_x = 16, local_size_y = 16, local_size_z = 1) in;
const vec2 workGroupsRender = vec2(1.0, 1.0);

uniform layout(rgba16) restrict writeonly image2D colorimg0;
uniform sampler2D blendWeightS, tempColS;

#include "/lib/view_size.glsl"
#include "/lib/srgb.glsl"

shared f16vec3[gl_WorkGroupSize.x + 2][gl_WorkGroupSize.y + 2] neighborhood;

void init_offset_weight(f16vec4 a, out vec4 blend_offset, out f16vec2 blend_weight) {
	immut bool h = max(a.x, a.z) > max(a.y, a.w);

	blend_offset = h ? vec4(a.x, 0.0, a.z, 0.0) : vec4(0.0, a.y, 0.0, a.w);

	blend_weight = h ? a.xz : a.yw;
	blend_weight /= dot(blend_weight, f16vec2(1.0));
}

f16vec3 blend_color(f16vec2 blend_weight, vec3 color_0, vec3 color_1) {
	return blend_weight.x * f16vec3(color_0) + blend_weight.y * f16vec3(color_1);
}

void main() {
	immut i16vec2 texel = i16vec2(gl_GlobalInvocationID.xy);

	/*
		SMAA Neighborhood Blending
		https://github.com/Luracasmus/smaa-mc/blob/main/shaders/composite2.csh

		Copyright (C) 2013 Jorge Jimenez (jorge@iryoku.com)
		Copyright (C) 2013 Jose I. Echevarria (joseignacioechevarria@gmail.com)
		Copyright (C) 2013 Belen Masia (bmasia@unizar.es)
		Copyright (C) 2013 Fernando Navarro (fernandn@microsoft.com)
		Copyright (C) 2013 Diego Gutierrez (diegog@unizar.es)
		Copyright (C) 2024 Luracasmus

		Permission is hereby granted, free of charge, to any person obtaining a copy
		this software and associated documentation files (the "Software"), to deal in
		the Software without restriction, including without limitation the rights to
		use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
		the Software, and to permit persons to whom the Software is furnished to do so,
		subject to the following conditions:

		The above copyright notice and this permission notice shall be included in all
		copies or substantial portions of the Software. As clarification, there is no
		requirement that the copyright notice and permission be included in binary
		distributions of the Software.

		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
		IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
		FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
		COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
		IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
		CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
	*/

	f16vec4 a = f16vec4(
		texelFetchOffset(blendWeightS, texel, 0, ivec2(1, 0)).w,
		texelFetchOffset(blendWeightS, texel, 0, ivec2(0, 1)).y,
		texelFetch(blendWeightS, texel, 0).zx
	);

	immut vec2 texel_size = 1.0 / vec2(view_size());
	immut vec2 coord = fma(vec2(texel), texel_size, 0.5 * texel_size);
	immut i8vec2 neighborhood_pos = i8vec2(gl_LocalInvocationID.xy) + int8_t(1);

	f16vec3 color;
	if (dot(vec4(a), vec4(1.0)) < 1.0e-5) {
		color = f16vec3(texelFetch(tempColS, texel, 0).rgb);
	} else {
		vec4 blend_offset; f16vec2 blend_weight;
		init_offset_weight(a, blend_offset, blend_weight);

		color = blend_color(
			blend_weight,
			textureLod(tempColS, fma(blend_offset.xy, texel_size, coord), 0.0).rgb,
			textureLod(tempColS, fma(blend_offset.zw, -texel_size, coord), 0.0).rgb
		);
	}

	neighborhood[neighborhood_pos.x][neighborhood_pos.y] = color;

	// BORDER_OP(offset)
	/*
		border_offset = i8vec2(offset);

		f16vec4 offset_a = f16vec4(
			texelFetchOffset(blendWeightS, texel, 0, offset + ivec2(1, 0)).w,
			texelFetchOffset(blendWeightS, texel, 0, offset + ivec2(0, 1)).y,
			texelFetchOffset(blendWeightS, texel, 0, offset).zx
		);

		if (dot(vec4(offset_a), vec4(1.0)) < 1.0e-5) {
			offset_color = f16vec3(texelFetchOffset(tempColS, texel, 0, offset).rgb);
		} else {
			vec4 blend_offset; f16vec2 blend_weight;
			init_offset_weight(offset_a, blend_offset, blend_weight);

			offset_color = blend_color(
				blend_weight,
				textureLodOffset(tempColS, fma(blend_offset.xy, texel_size, coord), 0.0, offset).rgb,
				textureLodOffset(tempColS, fma(blend_offset.zw, -texel_size, coord), 0.0, offset).rgb
			);
		}
	*/

	i8vec2 border_offset;
	f16vec3 offset_color = f16vec3(0.0);
	#define BORDER_OP(offset) border_offset=i8vec2(offset);f16vec4 offset_a=f16vec4(texelFetchOffset(blendWeightS,texel,0,offset+ivec2(1,0)).w,texelFetchOffset(blendWeightS,texel,0,offset+ivec2(0,1)).y,texelFetchOffset(blendWeightS,texel,0,offset).zx);if(dot(vec4(offset_a),vec4(1.0))<1.0e-5){offset_color=f16vec3(texelFetchOffset(tempColS,texel,0,offset).rgb);}else{vec4 blend_offset;f16vec2 blend_weight;init_offset_weight(offset_a,blend_offset,blend_weight);offset_color=blend_color(blend_weight,textureLodOffset(tempColS,fma(blend_offset.xy, texel_size, coord),0.0,offset).rgb,textureLodOffset(tempColS,fma(blend_offset.zw,-texel_size,coord),0.0,offset).rgb);};
	#define NON_BORDER_OP border_offset = i8vec2(0);
	#include "/lib/neighborhood/border_cornerless.glsl"

	if (border_offset != i8vec2(0)) {
		immut i8vec2 border_pos = neighborhood_pos + border_offset;
		neighborhood[border_pos.x][border_pos.y] = offset_color;
	}

	barrier();

	/*
		RCA Sharpening
		https://github.com/GPUOpen-LibrariesAndSDKs/FidelityFX-SDK/blob/main/sdk/include/FidelityFX/gpu/fsr1/ffx_fsr1.h#L684

		Copyright (C) 2024 Advanced Micro Devices, Inc.

		Permission is hereby granted, free of charge, to any person obtaining a copy
		of this software and associated documentation files(the "Software"), to deal
		in the Software without restriction, including without limitation the rights
		to use, copy, modify, merge, publish, distribute, sublicense, and /or sell
		copies of the Software, and to permit persons to whom the Software is
		furnished to do so, subject to the following conditions :

		The above copyright notice and this permission notice shall be included in
		all copies or substantial portions of the Software.

		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
		IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
		FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
		AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
		LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
		OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
		THE SOFTWARE.
	*/

	immut f16vec3 down = neighborhood[neighborhood_pos.x][neighborhood_pos.y - int8_t(1)];
	immut f16vec3 left = neighborhood[neighborhood_pos.x - int8_t(1)][neighborhood_pos.y];
	immut f16vec3 right = neighborhood[neighborhood_pos.x + int8_t(1)][neighborhood_pos.y];
	immut f16vec3 up = neighborhood[neighborhood_pos.x][neighborhood_pos.y + int8_t(1)];

	immut f16vec3 mn4 = min(down, min3(left, right, up));
	immut f16vec3 mx4 = max(down, max3(left, right, up));

	immut f16vec3 lobe_rgb = max(mn4 / (mx4 * float16_t(-4.0)), (float16_t(1.0) - mx4) / fma(mn4, f16vec3(4.0), f16vec3(-4.0)));
	float16_t lobe = clamp(max3(lobe_rgb.r, lobe_rgb.g, lobe_rgb.b), float16_t(-0.1875), float16_t(0.0)) * float16_t(SHARPNESS);

	// Denoise

	immut f16vec4 l = fma(
		f16vec4(down.b, left.b, right.b, up.b) + f16vec4(down.r, left.r, right.r, up.r),
		f16vec4(0.5),
		f16vec4(down.g, left.g, right.g, up.g) //not much, what's up with you?
	);

	immut float16_t noise = abs(dot(f16vec4(l.x, l.y, l.z, l.w), f16vec4(0.25)) - float16_t(0.5) * (color.b + color.r) - color.g);

	lobe *= fma(saturate(noise / (max(l.x, max3(l.y, l.z, l.w)) - min(l.x, min3(l.y, l.z, l.w)))), float16_t(-0.5), float16_t(1.0));

	color = fma(down + left + right + up, lobe.xxx, color) / fma(lobe, float16_t(4.0), float16_t(1.0));

	// color.rb += vec2(equal(gl_LocalInvocationID.xy, uvec2(0u)));

	imageStore(colorimg0, texel, vec4(srgb(color), 0.0));
}