#include "/prelude/core.glsl"

/* RENDERTARGETS: 1,2,3,4 */
layout(location = 0) out vec4 colortex1;
layout(location = 1) out vec4 colortex2;
layout(location = 2) out vec2 colortex3;
layout(location = 3) out vec3 colortex4;

#ifdef ALPHA_CHECK
	layout(depth_greater) out float gl_FragDepth;

	uniform float alphaTestRef;
#else
	layout(depth_unchanged) out float gl_FragDepth;
#endif

uniform sampler2D gtexture;

#ifdef NO_NORMAL
	uniform mat4 gbufferModelViewInverse;
#else
	#include "/lib/tbn/fsh.glsl"
#endif

#ifdef TERRAIN
	uniform ivec2 atlasSize;
#endif

#include "/lib/octa_normal.glsl"
#include "/lib/luminance.glsl"
#include "/lib/material/specular.glsl"
#include "/lib/material/normal.glsl"
#include "/lib/srgb.glsl"

in VertexData {
	layout(location = 2, component = 0) vec2 coord;
	layout(location = 2, component = 2) vec2 light;
	layout(location = 3, component = 0) vec3 tint;
	layout(location = 3, component = 3) float ao;
	layout(location = 4, component = 0) vec3 s_screen;

	#if !(NORMALS == 1 && defined MC_NORMAL_MAP)
		layout(location = 5, component = 0) flat uint mid_coord;
		layout(location = 5, component = 1) flat uint face_tex_size;
	#endif

	#ifdef TERRAIN
		layout(location = 5, component = 2) flat uint emission;

		#if !(SM && defined MC_SPECULAR_MAP)
			layout(location = 1, component = 3) flat float avg_luma;
		#endif
	#endif
} v;

void main() {
	#ifdef TEX_ALPHA
		f16vec4 color = f16vec4(texture(gtexture, v.coord));
	#else
		f16vec3 color = f16vec3(texture(gtexture, v.coord).rgb);
	#endif

	#ifdef ALPHA_CHECK
		if (color.a < float16_t(alphaTestRef)) discard;
	#endif

	colortex4 = v.s_screen;

	immut f16vec3 tint = f16vec3(v.tint);
	color.rgb *= tint;

	#ifdef NO_NORMAL
		immut vec3 normal = normalize(mat3(gbufferModelViewInverse) * vec3(0.0, 0.0, 1.0));
		immut vec3 tangent = normalize(mat3(gbufferModelViewInverse) * vec3(0.0, 1.0, 0.0));
		immut mat3 tbn = mat3(tangent, normalize(cross(tangent, normal)), normal);
	#else
		immut mat3 tbn = get_tbn();
	#endif

	#if !(NORMALS == 1 && defined MC_NORMAL_MAP) || !(SM && defined MC_SPECULAR_MAP)
		immut float16_t luma = luminance(color.rgb);
	#endif

	#if NORMALS == 1 && defined MC_NORMAL_MAP
		immut f16vec3 w_tex_normal = normalize(f16vec3(tbn * sample_normal(texture(normals, v.coord).rg)));
	#elif NORMALS == 2
		immut f16vec3 w_tex_normal = f16vec3(tbn[2]);
	#else
		#ifdef TERRAIN
			immut vec2 atlas = vec2(atlasSize);
		#else
			immut vec2 atlas = vec2(textureSize(gtexture, 0));
		#endif

		immut f16vec3 w_tex_normal = normalize(f16vec3(tbn * gen_normal(gtexture, tint, v.coord, v.mid_coord, v.face_tex_size, atlas, luma)));

		/*
		immut ivec2 half_texels = ivec2(
			v.face_tex_size & 65535u,
			bitfieldExtract(v.face_tex_size, 16, 16)
		) / 2 - 1;
		immut vec2 atlas_texel = 1.0 / atlas;
		immut vec2 half_size = vec2(half_texels) * atlas_texel;

		immut vec2 local_coord = v.coord - unpackUnorm2x16(v.mid_coord);
		color.rgb += vec4(
			local_coord.x > -half_size.x,
			local_coord.x < half_size.x,
			local_coord.y > -half_size.y,
			local_coord.y < half_size.y
		).rgb;
		*/
	#endif

	colortex2 = vec4(
		octa_encode(w_tex_normal),
		octa_encode(f16vec3(tbn[2]))
	);

	#if SM && defined MC_SPECULAR_MAP
		immut float16_t roughness = map_roughness(float16_t(texture(specular, v.coord).SM_CH));
	#else
		immut float16_t avg_luma =
			#ifdef TERRAIN
				float16_t(v.avg_luma);
			#else
				float16_t(0.8);
			#endif

		immut float16_t roughness = gen_roughness(luma, avg_luma);
	#endif

	color.rgb = linear(color.rgb);

	// todo!() emissive map here
	#ifdef TERRAIN
		// pack into sign bits
		color.rgb *= mix(f16vec3(-1.0), f16vec3(1.0), bvec3(
			v.emission & 1u,
			bitfieldExtract(v.emission, 1, 1),
			bitfieldExtract(v.emission, 2, 1)
		));
	#else
		color.rgb *= float16_t(-1.0);
	#endif

	colortex1 = vec4(color.rgb, roughness);

	immut float16_t in_ao = float16_t(v.ao);
	immut float16_t ao = mix(
		smoothstep(float16_t(0.05), float16_t(0.8), in_ao),
		float16_t(1.0),
		float16_t(0.25)
	) * mix(in_ao, float16_t(1.0), float16_t(0.75));

	colortex3 = ao * f16vec2(v.light.x, saturate(fma(in_ao, float16_t(AMBIENT), float16_t(v.light.y))));
}