#include "/prelude/core.glsl"

/* RENDERTARGETS: 1 */
layout(location = 0) out vec4 colortex1;

#ifdef ALPHA_CHECK
	layout(depth_greater) out float gl_FragDepth;

	uniform float alphaTestRef;
#else
	layout(depth_unchanged) out float gl_FragDepth;
#endif

uniform sampler2D gtexture;

#ifdef TERRAIN
	uniform ivec2 atlasSize;
#endif

#ifdef NO_NORMAL
	uniform mat4 gbufferModelViewInverse;
#else
	#include "/lib/tbn/fsh.glsl"
#endif

in VertexData {
	layout(location = 2, component = 0) vec3 tint;
	layout(location = 2, component = 3) float ao;
	layout(location = 3, component = 0) vec3 light;
	layout(location = 4, component = 0) vec3 view;
	layout(location = 5, component = 0) vec2 coord;

	#ifdef TERRAIN
		layout(location = 1, component = 3) flat float alpha;

		#if !(SM && defined MC_SPECULAR_MAP)
			layout(location = 6, component = 0) flat float avg_luma;
		#endif
	#endif

	#ifndef NETHER
		layout(location = 7, component = 0) vec3 s_screen;
	#endif

	#if !(NORMALS == 1 && defined MC_NORMAL_MAP)
		layout(location = 8, component = 0) flat uint mid_coord;
		layout(location = 8, component = 1) flat uint face_tex_size;
	#endif
} v;

#ifndef NETHER
	uniform vec3 shadowLightDirection;
	
	#include "/lib/skylight.glsl"
	#include "/lib/sm/distort.glsl"
	#include "/lib/sm/shadows.glsl"
#endif

#ifdef END
	uniform float frameTimeCounter;
	#include "/lib/prng/fast_rand.glsl"
#endif

#include "/lib/view_size.glsl"
#include "/lib/mmul.glsl"
#include "/lib/luminance.glsl"
#include "/lib/srgb.glsl"
#include "/lib/fog.glsl"
#include "/lib/material/specular.glsl"
#include "/lib/material/normal.glsl"

void main() {
	f16vec4 color = f16vec4(texture(gtexture, v.coord));

	#ifdef ALPHA_CHECK
		if (color.a < float16_t(alphaTestRef)) discard;
	#endif

	immut f16vec3 tint = f16vec3(v.tint);

	#ifdef TERRAIN
		color *= f16vec4(tint, v.alpha);
	#else
		color.rgb *= tint;
	#endif

	#ifdef NO_NORMAL
		immut vec3 normal = normalize(mat3(gbufferModelViewInverse) * vec3(0.0, 0.0, 1.0));
		immut vec3 tangent = normalize(mat3(gbufferModelViewInverse) * vec3(0.0, 1.0, 0.0));
		immut mat3 tbn = mat3(tangent, normalize(cross(tangent, normal)), normal);
	#else
		immut mat3 tbn = get_tbn();
	#endif

	#if SM && defined MC_SPECULAR_MAP
		immut float16_t roughness = map_roughness(float16_t(texture(specular, v.coord).SM_CH));
	#else
		immut float16_t avg_luma =
			#ifdef TERRAIN
				float16_t(v.avg_luma);
			#else
				float16_t(0.8);
			#endif

		immut float16_t roughness = gen_roughness(luminance(color.rgb), avg_luma);
	#endif

	#if NORMALS == 1 && defined MC_NORMAL_MAP
		immut f16vec3 v_tex_normal = normalize(f16vec3(tbn * sample_normal(texture(normals, v.coord).rg)));
	#elif NORMALS == 2
		immut f16vec3 v_tex_normal = f16vec3(tbn[2]);
	#else
		#ifdef TERRAIN
			immut vec2 atlas = vec2(atlasSize);
		#else
			immut vec2 atlas = vec2(textureSize(gtexture, 0));
		#endif

		immut f16vec3 v_tex_normal = normalize(f16vec3(tbn * gen_normal(gtexture, tint, v.coord, v.mid_coord, v.face_tex_size, atlas, luminance(color.rgb))));
	#endif

	color.rgb = linear(color.rgb);

	f16vec3 lighting = f16vec3(v.light);

	#ifndef NETHER
		immut f16vec3 skylight = skylight();

		#ifdef NO_NORMAL
			const float16_t face_n_dot_l = float16_t(1.0);
			const float16_t tex_n_dot_l = float16_t(1.0);
		#else
			immut float16_t face_n_dot_l = float16_t(dot(tbn[2], shadowLightDirection));
			immut float tex_n_dot_l = dot(v_tex_normal, shadowLightDirection);
		#endif

		#if SSS
			f16vec3 light = sample_shadow(v.s_screen);
		#endif

		if (face_n_dot_l > float16_t(0.01) && tex_n_dot_l > 0.0) {
			#if !SSS
				f16vec3 light = sample_shadow(v.s_screen);
			#endif

			if (dot(light, f16vec3(1.0)) > float16_t(0.0)) {
				immut f16vec3 shadow_light_dir = f16vec3(shadowLightDirection);

				immut f16vec2 specular_diffuse = brdf(face_n_dot_l, v_tex_normal, f16vec3(normalize(v.view)), shadow_light_dir, roughness);

				light *= float16_t(3.0) * (specular_diffuse.y + specular_diffuse.x / max(color.rgb, float16_t(1.0e-5)));

				lighting = fma(light, skylight, lighting);
			}
		}

		#if SSS
			else {
				lighting = fma(light, skylight, lighting);
			}
		#endif
	#endif

	color.rgb *= lighting;

	/*
	immut float solid_depth = texelFetch(depthtex1, ivec2(gl_FragCoord.xy), 0).r;

	if (solid_depth < 1.0) {
		immut vec3 solid_ndc = fma(vec3(gl_FragCoord.xy / view_size(), solid_depth), vec3(2.0), vec3(-1.0));
		immut vec3 solid_pe = mat3(gbufferModelViewInverse) * proj_inv(gbufferProjectionInverse, solid_ndc);
		immut float16_t fog = min(fog(solid_pe) + float16_t(1.0 - exp(-0.0125 / fogState.y * length(solid_pe))), float16_t(1.0)); // todo!() make this less cursed

		#if defined END || defined NETHER
			color.rgb = mix(color.rgb, color.rgb * linear(f16vec3(fogColor)), fog);
		#else
			immut vec3 n_pe = normalize(solid_pe);
			immut float16_t sky_fog = sky_fog(float16_t(n_pe.y));
			immut f16vec3 fog_col = sky(sky_fog, n_pe, mat3(gbufferModelViewInverse) * shadowLightDirection);
			color.rgb = mix(color.rgb, mix(color.rgb * fog_col, fog_col, fog), fog);
		#endif

		color.a = saturate(color.a + fog);
	} // todo!() self-colored fog should be based on the distance between the current surface and the solid one behind it, not the distance from the camera to the solid surface
	*/

	color.a *= float16_t(1.0) - edge_fog(v.view);

	colortex1 = color;
}